"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const common_1 = require("./phpunit/common");
const path_1 = require("path");
const helpers_1 = require("./helpers");
const vscode_1 = require("vscode");
class DecorateManager {
    constructor(client, context, win = vscode_1.window) {
        this.client = client;
        this.context = context;
        this.win = win;
        this.styles = new Map();
        this.styles.set(common_1.Type.PASSED, this.passed());
        this.styles.set(common_1.Type.ERROR, this.error());
        this.styles.set(common_1.Type.WARNING, this.skipped());
        this.styles.set(common_1.Type.FAILURE, this.error());
        this.styles.set(common_1.Type.INCOMPLETE, this.incomplete());
        this.styles.set(common_1.Type.RISKY, this.risky());
        this.styles.set(common_1.Type.SKIPPED, this.skipped());
        this.styles.set(common_1.Type.FAILED, this.error());
    }
    decoratedGutter(editor, assertions) {
        this.clearDecoratedGutter(editor);
        for (const [type, decorationOptions] of this.groupBy(assertions)) {
            helpers_1.when(this.styles.get(type), (style) => {
                editor.setDecorations(style, decorationOptions);
            });
        }
        return this;
    }
    clearDecoratedGutter(editor) {
        for (const [type] of this.styles) {
            helpers_1.when(this.styles.get(type), (style) => {
                editor.setDecorations(style, []);
            });
        }
        return this;
    }
    listen() {
        this.client.onNotification('assertions', (params) => {
            helpers_1.when(this.win.activeTextEditor, (editor) => {
                if (editor.document.uri.toString() === params.uri) {
                    this.decoratedGutter(editor, params.assertions);
                }
            });
        });
        this.win.onDidChangeActiveTextEditor((editor) => {
            helpers_1.when(editor, (editor) => {
                this.client.sendRequest('assertions', {
                    uri: editor.document.uri.toString(),
                });
            });
        });
    }
    groupBy(assertions) {
        return assertions.reduce((groups, assertion) => {
            const test = assertion.related;
            const group = groups.get(test.type) || [];
            const { start, end } = assertion.range;
            const related = assertion.related;
            group.push({
                range: new vscode_1.Range(start.line, start.character, end.line, end.character),
                hoverMessage: related.fault ? related.fault.message : '',
            });
            groups.set(test.type, group);
            return groups;
        }, new Map());
    }
    passed() {
        return this.createTextEditorDecorationType('success.svg', '#62b455');
    }
    error() {
        return this.createTextEditorDecorationType('danger.svg', '#fe536a');
    }
    risky() {
        return this.createTextEditorDecorationType('danger-light.svg', '#ffa0a0');
    }
    skipped() {
        return this.createTextEditorDecorationType('warning.svg', '#d2a032');
    }
    incomplete() {
        return this.skipped();
    }
    createTextEditorDecorationType(image, color) {
        return this.win.createTextEditorDecorationType({
            overviewRulerColor: color,
            overviewRulerLane: vscode_1.OverviewRulerLane.Left,
            gutterIconPath: this.gutterIconPath(image),
        });
    }
    gutterIconPath(img) {
        return path_1.resolve(this.context.extensionPath, 'images', img);
    }
}
exports.DecorateManager = DecorateManager;
//# sourceMappingURL=decorate-manage.js.map