"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_1 = require("vscode-languageserver");
const phpunit_1 = require("./phpunit");
const helpers_1 = require("./helpers");
const filesystem_1 = require("./filesystem");
class LangServer {
    constructor(connection = vscode_languageserver_1.createConnection(vscode_languageserver_1.ProposedFeatures.all), documents = new vscode_languageserver_1.TextDocuments(), phpUnit = new phpunit_1.PhpUnit(), files = new filesystem_1.Filesystem()) {
        this.connection = connection;
        this.documents = documents;
        this.phpUnit = phpUnit;
        this.files = files;
    }
    init() {
        this.documents.listen(this.connection);
        this.connection.onInitialize(this.onInitialize.bind(this));
        this.connection.onDidChangeConfiguration(this.onDidChangeConfiguration.bind(this));
        this.connection.onCodeLens(this.onCodeLens.bind(this));
        this.connection.onExecuteCommand(this.onExecuteCommand.bind(this));
        this.connection.onRequest('assertions', (params) => {
            this.sendAssertionNotification(params.uri);
        });
        this.connection.onDocumentSymbol(this.onDocumentSymbol.bind(this));
        return this;
    }
    listen() {
        this.connection.listen();
    }
    onInitialize() {
        return {
            capabilities: {
                // Tell the client that the server works in FULL text document sync mode
                textDocumentSync: this.documents.syncKind,
                // Tell the client that the server support code complete
                // completionProvider: {
                //     resolveProvider: true,
                // },
                codeLensProvider: {
                    resolveProvider: true,
                },
                documentSymbolProvider: true,
                executeCommandProvider: {
                    commands: [
                        'phpunit.test',
                        'phpunit.test.file',
                        'phpunit.test.suite',
                        'phpunit.test.nearest',
                        'phpunit.test.last',
                    ],
                },
            },
        };
    }
    onDidChangeConfiguration(change) {
        const settings = change.settings;
        this.phpUnit.setBinary(settings.phpunit.execPath).setDefault(settings.phpunit.args);
    }
    onCodeLens(params) {
        const textDocument = this.documents.get(params.textDocument.uri);
        return this.phpUnit.getCodeLens(textDocument.getText(), textDocument.uri);
    }
    onExecuteCommand(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const p = params.arguments || [];
            const uri = p[0] || '';
            const path = p[1] || '';
            const args = p[2] || [];
            this.connection.sendNotification('running');
            (yield this.executeCommand(params.command, path, args)).sendDiagnostics().sendAssertionNotification(uri);
            this.connection.sendNotification('done', this.phpUnit.getState());
            this.connection.console.log(this.phpUnit.getOutput());
        });
    }
    onDocumentSymbol(params) {
        const textDocument = this.documents.get(params.textDocument.uri);
        return this.phpUnit.getDocumentSymbols(textDocument.getText(), textDocument.uri);
    }
    sendAssertionNotification(pathOrUri) {
        return helpers_1.tap(this, () => {
            const uri = this.files.uri(pathOrUri);
            this.connection.sendNotification('assertions', {
                uri: uri,
                assertions: this.phpUnit.getAssertions(uri),
            });
        });
    }
    sendDiagnostics() {
        return helpers_1.tap(this, () => {
            this.phpUnit.getDiagnoics().forEach((diagnostics, uri) => {
                this.connection.sendDiagnostics({
                    uri,
                    diagnostics,
                });
            });
        });
    }
    executeCommand(command, path, args) {
        return __awaiter(this, void 0, void 0, function* () {
            switch (command) {
                case 'phpunit.test.nearest':
                    yield this.phpUnit.runNearest(path, args);
                    break;
                case 'phpunit.test.last':
                    yield this.phpUnit.runLast(path, args);
                    break;
                default:
                    yield this.phpUnit.run(path, args);
                    break;
            }
            return this;
        });
    }
}
exports.LangServer = LangServer;
//# sourceMappingURL=lang-server.js.map