"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const common_1 = require("./common");
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const filesystem_1 = require("../filesystem");
const helpers_1 = require("../helpers");
class Collection {
    constructor(files = new filesystem_1.Filesystem()) {
        this.files = files;
        this.errorTypes = [common_1.Type.ERROR, common_1.Type.FAILED, common_1.Type.FAILURE, common_1.Type.RISKY];
        this.items = new Map();
    }
    put(tests) {
        const groups = helpers_1.groupBy(tests, 'uri');
        for (const key of groups.keys()) {
            this.items.set(key, this.merge(this.asArray(this.items.get(key)), this.asArray(groups.get(key))));
        }
        return this;
    }
    get(uri) {
        return this.asArray(this.items.get(this.files.uri(uri)));
    }
    forEach(callback) {
        this.items.forEach((tests, uri) => {
            callback(tests, uri);
        });
        return this;
    }
    asTestNodes(uri) {
        return this.asArray(this.asAssertions().get(uri)).map((assertion) => {
            return Object.assign({}, this.cloneTest(assertion.related), {
                range: assertion.range,
            });
        });
    }
    asDiagnoics() {
        return helpers_1.tap(new Map(), (map) => {
            this.forEach((tests, uri) => {
                map.set(uri, tests
                    .map((test) => this.cloneTest(test, true))
                    .filter((test) => this.filterDetails(test))
                    .filter(this.filterByType.bind(this))
                    .map((test) => this.transformToDiagonstic(test)));
            });
        });
    }
    asAssertions(keepDetails = false) {
        const assertions = [];
        this.forEach((tests) => {
            tests.forEach((test) => {
                const details = this.getDetails(test.fault);
                const related = this.cloneTest(test, keepDetails);
                assertions.push({
                    uri: test.uri,
                    range: test.range,
                    related: related,
                });
                details.forEach((detail) => {
                    assertions.push({
                        uri: detail.uri,
                        range: detail.range,
                        related: related,
                    });
                });
            });
        });
        return helpers_1.groupBy(assertions, 'uri');
    }
    asState() {
        const state = {
            failed: 0,
            warning: 0,
            passed: 0,
        };
        this.forEach((tests) => {
            tests.forEach((test) => {
                if ([common_1.Type.ERROR, common_1.Type.FAILURE, common_1.Type.FAILED].indexOf(test.type) !== -1) {
                    state.failed++;
                }
                else if ([common_1.Type.INCOMPLETE, common_1.Type.RISKY, common_1.Type.SKIPPED, common_1.Type.WARNING].indexOf(test.type) !== -1) {
                    state.warning++;
                }
                else {
                    state.passed++;
                }
            });
        });
        return state;
    }
    transformToDiagonstic(test) {
        return {
            severity: test.type === common_1.Type.RISKY ? vscode_languageserver_types_1.DiagnosticSeverity.Warning : vscode_languageserver_types_1.DiagnosticSeverity.Error,
            range: test.range,
            message: test.fault ? test.fault.message : '',
            relatedInformation: this.transformToRelatedInformation(test),
            source: 'PHPUnit',
        };
    }
    transformToRelatedInformation(test) {
        return helpers_1.when(test.fault, (fault) => this.getDetails(fault).map((detail) => ({
            location: detail,
            message: fault.message,
        })), []);
    }
    filterByType(test) {
        return this.errorTypes.indexOf(test.type) !== -1;
    }
    merge(oldTests, newTests) {
        const merged = oldTests
            .filter((oldTest) => {
            for (const newTest of newTests) {
                if (oldTest.uri === newTest.uri && oldTest.name === newTest.name) {
                    return false;
                }
            }
            return true;
        })
            .concat(newTests);
        merged.sort((a, b) => {
            return a.range.start.line > b.range.start.line ? 1 : -1;
        });
        return merged;
    }
    cloneTest(test, keepDetails = false) {
        return helpers_1.tap({
            name: test.name,
            class: test.class,
            classname: test.classname,
            uri: test.uri,
            range: test.range,
            time: test.time,
            type: test.type,
        }, (related) => helpers_1.when(test.fault, (fault) => {
            related.fault = {
                type: fault.type,
                message: fault.message,
            };
            if (keepDetails === true) {
                related.fault.details = fault.details;
            }
        }));
    }
    filterDetails(test) {
        return helpers_1.tap(test, (test) => {
            helpers_1.when(test.fault, (fault) => {
                const details = this.getDetails(fault);
                const current = helpers_1.when(details, (details) => details.find((detail) => this.isSameLine(detail, test)));
                helpers_1.when(current, (current) => {
                    test.uri = current.uri;
                    test.range = current.range;
                    fault.details = details.filter((detail) => this.isSameLine(detail, current));
                });
            });
        });
    }
    isSameLine(source, target) {
        return source.uri === target.uri && target.range.start.line !== source.range.start.line;
    }
    getDetails(fault) {
        return fault && fault.details ? fault.details : [];
    }
    asArray(item) {
        return !item ? [] : item;
    }
}
exports.Collection = Collection;
//# sourceMappingURL=collection.js.map