"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const he_1 = require("he");
const common_1 = require("./common");
const filesystem_1 = require("../filesystem");
const helpers_1 = require("../helpers");
const textline_range_1 = require("./textline-range");
const parse = require('fast-xml-parser').parse;
class JUnit {
    constructor(files = new filesystem_1.Filesystem(), textLineRange = new textline_range_1.TextlineRange()) {
        this.files = files;
        this.textLineRange = textLineRange;
        this.pathPattern = /(.*):(\d+)$/;
    }
    parseFile(path) {
        return __awaiter(this, void 0, void 0, function* () {
            return path && (yield this.files.exists(path))
                ? helpers_1.tap(yield this.parse(yield this.files.get(path)), () => {
                    this.files.unlink(path);
                })
                : [];
        });
    }
    parse(code) {
        return __awaiter(this, void 0, void 0, function* () {
            return helpers_1.tap(yield this.getTests(this.getNodes(parse(code, {
                attributeNamePrefix: '_',
                ignoreAttributes: false,
                ignoreNameSpace: false,
                parseNodeValue: true,
                parseAttributeValue: true,
                trimValues: true,
                textNodeName: '__text',
            }))), () => this.textLineRange.clear());
        });
    }
    getSuites(node) {
        const suite = this.getSuite(node);
        return suite instanceof Array
            ? suite.reduce((suites, suite) => {
                return suites.concat(this.getSuites(suite));
            }, [])
            : [suite];
    }
    getSuite(node) {
        return helpers_1.when(node.testsuite, (testsuite) => {
            while (testsuite.testsuite) {
                testsuite = testsuite.testsuite;
            }
            return testsuite;
        }, node);
    }
    getNodes(node) {
        return this.getSuites(node.testsuites).reduce((tests, suite) => {
            return tests.concat(suite.testcase);
        }, []);
    }
    getTests(nodes) {
        return Promise.all(nodes.map(this.parseTest.bind(this)));
    }
    parseTest(node) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.parseFault(Object.assign(yield this.createLocation(node._file, parseInt(node._line, 10) || 1), {
                name: node._name || '',
                class: node._class,
                classname: node._classname || '',
                time: parseFloat(node._time) || 0,
                type: common_1.Type.PASSED,
            }), node);
        });
    }
    parseFault(test, node) {
        return helpers_1.when(this.getFaultNode(node), (fault) => __awaiter(this, void 0, void 0, function* () {
            const details = yield this.parseDetails(fault);
            const message = this.parseMessage(fault);
            return Object.assign(test, {
                type: fault.type,
                fault: {
                    type: fault._type || '',
                    message: message,
                    details: details,
                },
            });
        }), test);
    }
    getFaultNode(node) {
        const keys = Object.keys(node);
        if (keys.indexOf('error') !== -1) {
            return helpers_1.tap(node.error, (fault) => (fault.type = this.parseErrorType(fault)));
        }
        if (keys.indexOf('warning') !== -1) {
            return helpers_1.tap(node.warning, (fault) => (fault.type = common_1.Type.WARNING));
        }
        if (keys.indexOf('failure') !== -1) {
            return helpers_1.tap(node.failure, (fault) => (fault.type = common_1.Type.FAILURE));
        }
        if (keys.indexOf('skipped') !== -1) {
            return {
                type: common_1.Type.SKIPPED,
                _type: 'PHPUnit\\Framework\\SkippedTestError',
                __text: 'Skipped Test',
            };
        }
        if (keys.indexOf('incomplete') !== -1) {
            return {
                type: common_1.Type.INCOMPLETE,
                _type: 'PHPUnit\\Framework\\IncompleteTestError',
                __text: 'Incomplete Test',
            };
        }
        return;
    }
    parseErrorType(fault) {
        const type = fault._type.toLowerCase();
        return ([common_1.Type.WARNING, common_1.Type.FAILURE, common_1.Type.INCOMPLETE, common_1.Type.RISKY, common_1.Type.SKIPPED, common_1.Type.FAILED].find(errorType => type.indexOf(errorType) !== -1) || common_1.Type.ERROR);
    }
    parseDetails(fault) {
        return __awaiter(this, void 0, void 0, function* () {
            return Promise.all(fault.__text
                .split(/\r?\n/)
                .map((line) => line.trim())
                .filter((line) => this.pathPattern.test(line))
                .map((detail) => __awaiter(this, void 0, void 0, function* () {
                const [, file, line] = detail.match(this.pathPattern);
                return this.createLocation(file.trim(), parseInt(line, 10));
            })));
        });
    }
    parseMessage(fault) {
        const messages = fault.__text
            .replace(/\r?\n/g, '\n')
            .replace(/&#13;/g, '')
            .split('\n')
            .map((line) => line.replace(this.pathPattern, ''));
        return helpers_1.value(messages.slice(messages.length === 1 ? 0 : 1).join('\n'), (message) => {
            const type = fault._type || '';
            return (he_1.decode(/phpunit/i.test(type)
                ? message.replace(new RegExp(`^${type.replace(/\\/g, '\\\\')}:`), '')
                : message).trim() || type);
        });
    }
    createLocation(file, line) {
        return __awaiter(this, void 0, void 0, function* () {
            const uri = this.files.uri(file);
            return {
                uri: uri,
                range: yield this.textLineRange.create(uri, line - 1),
            };
        });
    }
}
exports.JUnit = JUnit;
//# sourceMappingURL=junit.js.map