"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const ast_1 = require("./ast");
const cli_1 = require("./cli");
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const collection_1 = require("./collection");
const filesystem_1 = require("./../filesystem");
const helpers_1 = require("../helpers");
const textline_range_1 = require("./textline-range");
class PhpUnit {
    constructor(cli = new cli_1.Cli(), collect = new collection_1.Collection(), ast = new ast_1.Ast(), files = new filesystem_1.Filesystem(), textlineRange = new textline_range_1.TextlineRange()) {
        this.cli = cli;
        this.collect = collect;
        this.ast = ast;
        this.files = files;
        this.textlineRange = textlineRange;
        this.output = '';
        this.lastCommand = {
            path: '',
            params: [],
        };
        this.cli.on('done', (output, tests, path, params) => {
            this.output = output;
            if (tests.length > 0) {
                this.collect.put(tests);
                this.lastCommand = { path, params };
            }
        });
    }
    setBinary(binary) {
        return helpers_1.tap(this, () => {
            this.cli.setBinary(binary);
        });
    }
    setDefault(args) {
        return helpers_1.tap(this, () => {
            this.cli.setDefault(args);
        });
    }
    getTestNodes(code, uri) {
        return this.ast.parse(code, this.files.uri(uri));
    }
    getCodeLens(code, uri) {
        const testNodes = this.getTestNodes(code, uri);
        return testNodes
            .concat(this.collect.asTestNodes(uri).filter((testNode) => {
            for (const node of testNodes) {
                if (node.uri === testNode.uri && testNode.range === testNode.range) {
                    return false;
                }
            }
            return true;
        }))
            .map((node) => this.asCodeLen(node, uri));
    }
    getDiagnoics() {
        return this.collect.asDiagnoics();
    }
    getAssertions(uri) {
        return this.collect.asAssertions().get(uri) || [];
    }
    getDocumentSymbols(code, uri) {
        return this.getTestNodes(code, uri).map((node) => this.asDocumentSymbol(node, uri));
    }
    getState() {
        return this.collect.asState();
    }
    getOutput() {
        return this.output;
    }
    run(path, params) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.cli.run(path, params);
            return this;
        });
    }
    runLast(path = '', params = []) {
        return __awaiter(this, void 0, void 0, function* () {
            path = this.lastCommand.path || path;
            params = this.lastCommand.params || params;
            return yield this.run(path, params);
        });
    }
    runNearest(path, params = []) {
        return __awaiter(this, void 0, void 0, function* () {
            return helpers_1.tap(yield this.run(path, this.asMethodFilter(yield this.textlineRange.findMethod(path, parseInt(params[0], 10)))), () => this.textlineRange.clear());
        });
    }
    asCodeLen(node, uri) {
        return {
            range: node.range,
            command: helpers_1.when(node.class === node.name, () => {
                return {
                    title: 'Run Test',
                    command: 'phpunit.test.file',
                    arguments: [uri, this.files.normalizePath(node.uri), []],
                };
            }, () => {
                return {
                    title: 'Run Test',
                    command: 'phpunit.test',
                    arguments: [uri, this.files.normalizePath(node.uri), this.asMethodFilter(node.name)],
                };
            }),
            data: {
                textDocument: {
                    uri: uri,
                },
            },
        };
    }
    asMethodFilter(method) {
        return method ? ['--filter', `^.*::${method}( with data set .*)?$`] : [];
    }
    asDocumentSymbol(node, uri) {
        return vscode_languageserver_types_1.SymbolInformation.create(node.name.replace(/.*\\/g, ''), node.class === node.name ? vscode_languageserver_types_1.SymbolKind.Class : vscode_languageserver_types_1.SymbolKind.Method, node.range, uri);
    }
}
exports.PhpUnit = PhpUnit;
//# sourceMappingURL=phpunit.js.map